/*
 * Generator.cpp
 *
 *  Created on: 18 окт. 2016 г.
 *      Author: ryutin
 */

#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <stdio.h>
#include <cassert>
#include <time.h>
#include <ctime>

#include "../inc/I.h"
#include "../inc/PI.h"
#include "../inc/NDarray.h"
#include "../inc/NDgrid.h"
#include "../inc/iFunc.h"
#include "../inc/newrand.h"
#include "../inc/Generator.h"


///////////////////////////////////////////////////////////////////
ExDiff::Generator::Generator (const ExDiff::iFunc& FUN_):
	    FUN(FUN_),
        FUNG(CalcFUNG(FUN_)),
		FUNGTOT(1.0)
{
              assert(FUN.GetMaxFun() > 0.0);
			  assert(!FUNG.empty());
			  assert(FUNGTOT > 0.0);
// initialization	
			  std::vector<double> f = FUNG[FUNG.size()-1].GetTab().GetValues();
			  std::vector<double> x = FUNG[FUNG.size()-1].GetGrid().GetValues();              	
              FUNGTOT = IntVec(x,f);
			  
};
//
ExDiff::Generator::~Generator(){
	
};
// print information
void ExDiff::Generator::printC(){
	std::cout<< "GENERATOR printC() ============================================= " <<std::endl;	
    std::cout<< "FUN printC() ============================================= " <<std::endl;	
	FUN.printC();
	std::cout<< "FUNGS printC() ============================================= " <<std::endl;	
	for(std::size_t i = 0; i < FUN.GetDimensions().size(); i++){
	 FUNG[i].printC();	
	}
	std::cout<< "FUNGTOT = " << FUNGTOT <<std::endl;	 
};
// recursion to n-1 iFunc
ExDiff::iFunc ExDiff::Generator::CalcMinus(const ExDiff::iFunc& FUN_){
// check dimensions.size()>1 when recurrently calc. minus funs !	
// extract vectors to initialize new iFunc
        std::vector<size_t> in_dimensions = FUN_.GetTab().GetDimensions(); 
        std::vector<size_t> out_dimensions = in_dimensions;
		out_dimensions.pop_back();	
// compute total sizes for Tab and Grid general vectors
        std::size_t ntab = FUN_.GetTab().computeTotalSize(out_dimensions);
		std::size_t ngrid = FUN_.GetGrid().computeTotalSize(out_dimensions);
// new iFunc initialization		
		ExDiff::NDarray<double> TAB_(out_dimensions);
		ExDiff::NDgrid<double> GRID_(out_dimensions);
// fill in Tab and Grid Tab.GetValues()[i] and Grid.GetValues()[i] !!! check !!!
 		for(std::size_t i = 0; i < ngrid; i++){
		  GRID_.GetValues()[i] = FUN_.GetGrid().GetValues()[i];	
		}
		
		double res = 0.0;
		std::size_t nlast = in_dimensions.size()-1;
		std::vector<size_t> out_indexes_1 = out_dimensions;
		std::vector<size_t> out_indexes_2 = out_dimensions;
		
		for(std::size_t i = 0; i < ntab; i++){
		  res = 0.0;
          out_indexes_1 = TAB_.computeIndexes(i);
		  out_indexes_2 = out_indexes_1;
// 	  
          for(std::size_t j = 0; j < in_dimensions[in_dimensions.size()-1]-1; j++){
			  out_indexes_1.push_back(j);
			  out_indexes_2.push_back(j+1);
			  res +=0.5*(FUN_.GetTab().get(out_indexes_1)+FUN_.GetTab().get(out_indexes_2))*(FUN_.GetGrid().GetVar(nlast,j+1)-FUN_.GetGrid().GetVar(nlast,j));
			  out_indexes_1.pop_back();
			  out_indexes_2.pop_back();
		  }
 		  TAB_.GetValues()[i] = res;
		}
		
		ExDiff::iFunc MinusFUN_(TAB_,GRID_);
		return MinusFUN_;		
	};
// array of iFunc for generation	
std::vector<ExDiff::iFunc> ExDiff::Generator::CalcFUNG(const ExDiff::iFunc& FUN_){
	  std::vector<ExDiff::iFunc> FUNG_;
	  FUNG_.clear();	 
	  FUNG_.push_back(FUN_);	 
	  for(std::size_t i = 1 ; i < FUN_.GetDimensions().size(); i++){
       	FUNG_.push_back(CalcMinus(FUNG_[i-1]));       		
	  }
	  return FUNG_;
};
// aux for algorithm
std::vector<double> ExDiff::Generator::FGenVar(const std::vector<double>& x_, const std::vector<double>& f_){

//        std::cout << "GENVAR BEGIN ++++++++++++++++++++++++++++++++++++++" << std::endl;	
	
	 std::vector<double> F_ = {0.0};
	 double r = 0.0;
	 std::size_t i = 0;
	 for(i = 0; i < x_.size() - 1; i++){
		 r += (f_[i+1]+f_[i])*(x_[i+1]-x_[i])*0.5;
		 F_.push_back(r);
	 }
	 
	 return F_;
};
// faster generator
double ExDiff::Generator::GenVarFast1(const std::vector<double>& x_, const std::vector<double>& f_, const std::vector<double>& F_, const double& R_){
//        std::cout << "GENVAR BEGIN ++++++++++++++++++++++++++++++++++++++" << std::endl;	
	
	 double r = 0.0;
	 std::size_t i = 0;

	 double RR_ = R_*r;
	 std::size_t n = 0;
	 i = 0;
	 while(RR_ >= F_[i] && i < x_.size()){
			 n = i;
			 i++;
	 }
	 double a, b;
     a = 0.5*(f_[n+1]+f_[n])/(x_[n+1]-x_[n])/r;
     b = f_[n]/r;
     double xg_ = x_[n] + 0.5*(sqrt(b*b+4*a*(R_-F_[n]/r))-b)/a; 

//        std::cout << "GENVAR END ++++++++++++++++++++++++++++++++++++++" << std::endl;	
	 
	 return xg_;
};


// integrals of vector (Algorithm G of generation of one variable, R_ - random number [0,1]) 
double ExDiff::Generator::GenVar(const std::vector<double>& x_, const std::vector<double>& f_, const double& R_){

//        std::cout << "GENVAR BEGIN ++++++++++++++++++++++++++++++++++++++" << std::endl;	
	
	 std::vector<double> F_ = {0.0};
	 double r = 0.0;
	 std::size_t i = 0;
	 for(i = 0; i < x_.size() - 1; i++){
		 r += (f_[i+1]+f_[i])*(x_[i+1]-x_[i])*0.5;
		 F_.push_back(r);
	 }
	 double RR_ = R_*r;
	 std::size_t n = 0;
	 i = 0;
	 while(RR_ >= F_[i] && i < x_.size()){
			 n = i;
			 i++;
	 }
	 double a, b;
     a = 0.5*(f_[n+1]+f_[n])/(x_[n+1]-x_[n])/r;
     b = f_[n]/r;
     double xg_ = x_[n] + 0.5*(sqrt(b*b+4*a*(R_-F_[n]/r))-b)/a; 

//        std::cout << "GENVAR END ++++++++++++++++++++++++++++++++++++++" << std::endl;	
	 
	 return xg_;
};
// integral of the last recurrensy FUNG (total integral, sig_tot!) 
double ExDiff::Generator::IntVec(const std::vector<double>& x_, const std::vector<double>& f_){
	 double r = 0.0;
	 std::size_t i = 0;
	 for(i = 0; i < x_.size() - 1; i++){
		 r += (f_[i+1]+f_[i])*(x_[i+1]-x_[i])*0.5;
	 }
	 return r;
};
// generation
std::vector<double> ExDiff::Generator::Generate(){

//        std::cout << "GENERATE() BEGIN -------------------------------------" << std::endl;	

	std::vector<double> res, p_;
    res.clear(); 	
	std::vector<double> x_,f_;
    std::size_t i = 0;
    std::size_t j = 0;  
	std::size_t Dim = FUN.GetDimensions().size();
	ExDiff::NDgrid<double> Grid = FUN.GetGrid();
	std::vector<size_t> ND = FUN.GetDimensions();

	double xx = 0.0;
	double R_ = 0.0;

// time variables ////////////////////////////////
//    std::clock_t start;
//    double duration;
//////////////////////////////////////////////////

    if(Dim == 1){
		f_ = FUN.GetTab().GetValues();
		x_ = Grid.GetValues();	
		R_ = RNUM();
		res.push_back(GenVar(x_,f_,R_));		
	} else{		
////////////// for Dim > 1 start ///////////////////////////
	    for(i = 0; i < Dim; i++){
			
		    x_.clear();
	        f_.clear();
		    p_.clear(); 

//		    std::cout << "GENERATE() CYCLE BEGIN -------------------------------------" << std::endl;

      	    for(j = 0; j < ND[i]; j++){

// time for calculations start ///////////////////////////			
//			if(j == 0){							
//			  printf("clock start = %d\n", std::clock());
//			  start = std::clock();
//			}
//////////////////////////////////////////////////////////
		 
			   xx = Grid.GetVar(i,j); 
			   x_.push_back(xx);
			   res.push_back(xx);			
			   p_ = res;
			   res.pop_back();				
		       f_.push_back(FUNG[Dim-1-i].Calc(p_));

// time for calculations stop ///////////////////////////////////////////////			 				
//            if(j == ND[i]-1){ 
//			printf("clock stop = %d\n", std::clock());
//           duration = ( std::clock() - start ) / (double) CLOCKS_PER_SEC;
//            std::cout<<"duration = "<< duration << '\n';				
//			}		
/////////////////////////////////////////////////////////////////////////////			
			
		    } 
		 
//		    std::cout << "GENERATE() CYCLE END -------------------------------------" << std::endl;
		    R_ = RNUM();

//		 std::cout << ">>>>>>>>>>>> R = " << R_ << std::endl;
		    res.push_back(GenVar(x_,f_,R_));
		  
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!		 
		 
	    }
////////////// for Dim > 1 end ///////////////////////////	
	}	
//       std::cout << "GENERATE() END -------------------------------------" << std::endl;		
	
	return res;
};
// to define special random number 0->1
double ExDiff::Generator::RNUM(){
		double R_ = 0.000001*(rand()%1000000);
		return R_;		
};	